@@ -271,13 +271,21 @@ void disk_create_index(void) {
         block->drive = drive;
         block->partition = 0;
         block->first_sect = 0;
-        block->sect_count = drive_params.lba_count;
         block->max_partition = -1;
 
         if (!detect_sector_size(block)) {
+            pmm_free(block, sizeof(struct volume));
             continue;
         }
 
+        // Normalize sect_count to 512-byte sectors for consistency with partitions
+        // Preserve (uint64_t)-1 sentinel value (means "unknown size")
+        if (drive_params.lba_count == (uint64_t)-1 || drive_params.lba_count == 0) {
+            block->sect_count = (uint64_t)-1;
+        } else {
+            block->sect_count = drive_params.lba_count * (block->sector_size / 512);
+        }
+
         // Detect optical drives via DPTE ATAPI bit (bit 6) or sector size heuristic
         bool is_atapi = (dpte != NULL && (dpte->flags & (1 << 6)));
         block->is_optical = is_atapi || (block->sector_size == 2048 && is_removable);
@@ -386,6 +394,9 @@ static struct volume *pxe_from_efi_handle(EFI_HANDLE efi_handle) {
 
 #define UNIQUE_SECTOR_POOL_SIZE 65536
 static uint8_t *unique_sector_pool;
+static bool unique_sectors_calculated = false;
+
+static void find_unique_sectors(void);
 
 static struct volume *volume_by_unique_sector(void *b2b) {
     for (size_t i = 0; i < volume_index_i; i++) {
@@ -401,6 +412,23 @@ static struct volume *volume_by_unique_sector(void *b2b) {
     return NULL;
 }
 
+// Search for matching hash including invalidated volumes (for collision detection)
+static struct volume *volume_by_sector_hash(void *b2b) {
+    for (size_t i = 0; i < volume_index_i; i++) {
+        if (volume_index[i]->unique_sector_valid == false
+         && memcmp(volume_index[i]->unique_sector_b2b, (uint8_t[BLAKE2B_OUT_BYTES]){0}, BLAKE2B_OUT_BYTES) == 0) {
+            // Hash was never set, skip
+            continue;
+        }
+
+        if (memcmp(volume_index[i]->unique_sector_b2b, b2b, BLAKE2B_OUT_BYTES) == 0) {
+            return volume_index[i];
+        }
+    }
+
+    return NULL;
+}
+
 static bool is_efi_handle_to_skip(EFI_HANDLE efi_handle) {
     EFI_STATUS status;
 
@@ -555,7 +583,8 @@ static struct volume *volume_by_device_path(EFI_HANDLE query_handle) {
             while (!IsDevicePathEnd(qp)) {
                 if (qp->Type == MEDIA_DEVICE_PATH && qp->SubType == MEDIA_HARDDRIVE_DP) {
                     uint16_t len = DevicePathNodeLength(qp);
-                    if (len < sizeof(HARDDRIVE_DEVICE_PATH)) {
+                    // UEFI spec size is 42 bytes, but sizeof() may be larger due to padding
+                    if (len < 42) {
                         break;
                     }
                     HARDDRIVE_DEVICE_PATH *query_hd = (HARDDRIVE_DEVICE_PATH *)qp;
@@ -606,6 +635,13 @@ struct volume *disk_volume_from_efi_handle(EFI_HANDLE efi_handle) {
         return pxe_from_efi_handle(efi_handle);
     }
 
+    // Try device path matching first (primary method)
+    struct volume *ret = volume_by_device_path(efi_handle);
+    if (ret != NULL) {
+        return ret;
+    }
+
+    // Fallback to unique sector matching
     uint64_t bdev_size = ((uint64_t)block_io->Media->LastBlock + 1) * (uint64_t)block_io->Media->BlockSize;
     if (bdev_size >= UNIQUE_SECTOR_POOL_SIZE) {
         status = block_io->ReadBlocks(block_io, block_io->Media->MediaId,
@@ -613,20 +649,32 @@ struct volume *disk_volume_from_efi_handle(EFI_HANDLE efi_handle) {
                                       UNIQUE_SECTOR_POOL_SIZE,
                                       unique_sector_pool);
         if (status == 0) {
+            find_unique_sectors();
+
             uint8_t b2b[BLAKE2B_OUT_BYTES];
             blake2b(b2b, unique_sector_pool, UNIQUE_SECTOR_POOL_SIZE);
 
-            struct volume *ret = volume_by_unique_sector(b2b);
+            ret = volume_by_unique_sector(b2b);
             if (ret != NULL) {
-                return ret;
+                // Verify size, block size, and partition status match
+                if (block_io->Media->BlockSize == (uint32_t)ret->sector_size
+                 && bdev_size == ret->sect_count * 512
+                 && block_io->Media->LogicalPartition == (ret->partition != 0)) {
+                    return ret;
+                }
             }
         }
     }
 
-    return volume_by_device_path(efi_handle);
+    return NULL;
 }
 
 static void find_unique_sectors(void) {
+    if (unique_sectors_calculated) {
+        return;
+    }
+    unique_sectors_calculated = true;
+
     EFI_STATUS status;
 
     for (size_t i = 0; i < volume_index_i; i++) {
@@ -636,7 +684,8 @@ static void find_unique_sectors(void) {
 
         size_t first_sect = (volume_index[i]->first_sect * 512) / volume_index[i]->sector_size;
 
-        if (volume_index[i]->sect_count * volume_index[i]->sector_size < UNIQUE_SECTOR_POOL_SIZE) {
+        // sect_count is always in 512-byte sectors
+        if (volume_index[i]->sect_count * 512 < UNIQUE_SECTOR_POOL_SIZE) {
             continue;
         }
 
@@ -653,15 +702,21 @@ static void find_unique_sectors(void) {
         uint8_t b2b[BLAKE2B_OUT_BYTES];
         blake2b(b2b, unique_sector_pool, UNIQUE_SECTOR_POOL_SIZE);
 
-        struct volume *collision = volume_by_unique_sector(b2b);
+        // Check for collision BEFORE storing hash (so we don't find ourselves)
+        // This searches all volumes including previously invalidated ones
+        struct volume *collision = volume_by_sector_hash(b2b);
+
+        // Always store the hash so future volumes can detect collisions
+        memcpy(volume_index[i]->unique_sector_b2b, b2b, BLAKE2B_OUT_BYTES);
+
         if (collision == NULL) {
             volume_index[i]->unique_sector_valid = true;
-            memcpy(volume_index[i]->unique_sector_b2b, b2b, BLAKE2B_OUT_BYTES);
             continue;
         }
 
-        // Invalidate collision's unique sector
+        // Collision found - invalidate both volumes
         collision->unique_sector_valid = false;
+        volume_index[i]->unique_sector_valid = false;
     }
 }
 
@@ -728,6 +783,19 @@ fail:
         if (drive->Media->LogicalPartition)
             continue;
 
+        // Read test to ensure device is responsive (skipping this causes hangs on some systems)
+        status = drive->ReadBlocks(drive, drive->Media->MediaId, 0, 4096, unique_sector_pool);
+        if (status) {
+            continue;
+        }
+
+        if (drive->Media->BlockSize == 0) {
+            continue;
+        }
+        if (drive->Media->LastBlock == UINT64_MAX) {
+            continue;
+        }
+
         struct volume *block = ext_mem_alloc(sizeof(struct volume));
 
         bool is_optical = is_efi_handle_optical(handles[i]) ||
@@ -745,7 +813,8 @@ fail:
         block->partition = 0;
         block->sector_size = drive->Media->BlockSize;
         block->first_sect = 0;
-        block->sect_count = drive->Media->LastBlock + 1;
+        // Normalize sect_count to 512-byte sectors for consistency with partitions
+        block->sect_count = (drive->Media->LastBlock + 1) * (drive->Media->BlockSize / 512);
         block->max_partition = -1;
 
         if (drive->Revision >= EFI_BLOCK_IO_PROTOCOL_REVISION3) {
@@ -793,7 +862,6 @@ fail:
         }
     }
 
-    find_unique_sectors();
     find_part_handles(handles, handle_count);
 
     pmm_free(handles, handles_size);
