@@ -1,28 +1,18 @@
 #if defined (BIOS)
 
 #include <stdint.h>
-#include <stddef.h>
 #include <stdbool.h>
 #include <drivers/serial.h>
 #include <sys/cpu.h>
-#include <lib/misc.h>
-#include <lib/config.h>
 
 static bool serial_initialised = false;
-static uint32_t serial_baudrate;
+uint32_t serial_baudrate;
 
 static void serial_initialise(void) {
-    if (serial_initialised || config_ready == false) {
+    if (serial_initialised || !serial) {
         return;
     }
 
-    char *baudrate_s = config_get_value(NULL, 0, "SERIAL_BAUDRATE");
-    if (baudrate_s == NULL) {
-        serial_baudrate = 115200;
-    } else {
-        serial_baudrate = strtoui(baudrate_s, NULL, 10);
-    }
-
     // Init com1
     outb(0x3f8 + 3, 0x00);
     outb(0x3f8 + 1, 0x00);
