@@ -385,13 +385,11 @@ struct file_handle *iso9660_open(struct volume *vol, const char *path) {
             // Use the current directory's extent info
             // For root, this was set from ret->context->root
             // For subdirs, it was set from the last matched entry
-            if (first) {
-                // Still at root - return NULL as we can't return root as a file
-                pmm_free(ret, sizeof(struct iso9660_file_handle));
-                return NULL;
+            if (!first) {
+                pmm_free(current, current_size);
             }
-            // Otherwise next_sector/next_size already set from last entry
-            break;
+            pmm_free(ret, sizeof(struct iso9660_file_handle));
+            return NULL;
         }
 
         char *aux = filename;
@@ -490,18 +488,6 @@ struct file_handle *iso9660_open(struct volume *vol, const char *path) {
         }
     }
 
-    // Free the last directory buffer if we allocated one (not the cached root)
-    if (!first) {
-        pmm_free(current, current_size);
-    }
-
-    // Fallback path (trailing slash case) - create single extent
-    ret->extents = ext_mem_alloc(sizeof(struct iso9660_extent));
-    ret->extent_count = 1;
-    ret->total_size = next_size;
-    ret->extents[0].LBA = next_sector;
-    ret->extents[0].size = next_size;
-
 setup_handle:;
     struct file_handle *handle = ext_mem_alloc(sizeof(struct file_handle));
 
@@ -545,6 +531,10 @@ static void iso9660_read(struct file_handle *file, void *buf, uint64_t loc, uint
 
         extent_start = extent_end;
     }
+
+    if (count > 0) {
+        panic(false, "iso9660: read beyond end of file");
+    }
 }
 
 static void iso9660_close(struct file_handle *file) {
