@@ -292,7 +292,7 @@ int pit_sleep_and_quit_on_keypress(int seconds) {
 
     gBS->CreateEvent(EVT_TIMER, TPL_CALLBACK, NULL, NULL, &events[1]);
 
-    gBS->SetTimer(events[1], TimerRelative, 10000000 * seconds);
+    gBS->SetTimer(events[1], TimerRelative, (uint64_t)10000000 * seconds);
 
 again:
     memset(&kd, 0, sizeof(EFI_KEY_DATA));
@@ -323,6 +323,8 @@ again:
     }
 
     if (kd.Key.ScanCode == SCAN_ESC) {
+        gBS->CloseEvent(events[1]);
+
         gBS->CreateEvent(EVT_TIMER, TPL_CALLBACK, NULL, NULL, &events[1]);
 
         gBS->SetTimer(events[1], TimerRelative, 100000);
@@ -330,6 +332,7 @@ again:
         gBS->WaitForEvent(2, events, &which);
 
         if (which == 1) {
+            gBS->CloseEvent(events[1]);
             return GETCHAR_ESCAPE;
         }
 
@@ -339,11 +342,16 @@ again:
             status = exproto->ReadKeyStrokeEx(exproto, &kd);
         }
 
+        gBS->CloseEvent(events[1]);
+        gBS->CreateEvent(EVT_TIMER, TPL_CALLBACK, NULL, NULL, &events[1]);
+        gBS->SetTimer(events[1], TimerRelative, (uint64_t)10000000 * seconds);
+
         if (status != EFI_SUCCESS) {
             goto again;
         }
 
         if (kd.Key.UnicodeChar == '[') {
+            gBS->CloseEvent(events[1]);
             return input_sequence(!use_sproto, exproto, sproto);
         }
     }
