@@ -15,6 +15,7 @@
 #include <mm/pmm.h>
 #include <drivers/vbe.h>
 #include <drivers/vga_textmode.h>
+#include <drivers/serial.h>
 #include <protos/linux.h>
 #include <protos/chainload.h>
 #include <protos/multiboot1.h>
@@ -176,6 +177,10 @@ static void putchar_tokencol(int type, char c) {
 static bool editor_no_term_reset = false;
 
 char *config_entry_editor(const char *title, const char *orig_entry) {
+    if (terms[0]->cols < 40 || terms[0]->rows < 16) {
+        panic(false, "Terminal too small (minimum 40x16 required)");
+    }
+
     FOR_TERM(TERM->autoflush = false);
 
     FOR_TERM(TERM->cursor_enabled = true);
@@ -493,7 +498,7 @@ tab_part:
             if (cursor_offset) {
                 cursor_offset--;
         case GETCHAR_DELETE:
-                for (size_t i = cursor_offset; ; i++) {
+                for (size_t i = cursor_offset; i < buffer_len; i++) {
                     buffer[i] = buffer[i+1];
                     if (!buffer[i])
                         break;
@@ -803,6 +808,20 @@ noreturn void _menu(bool first_run) {
 #endif
         serial_str != NULL && strcmp(serial_str, "yes") == 0;
 
+#if defined (BIOS)
+    if (serial) {
+        char *baudrate_s = config_get_value(NULL, 0, "SERIAL_BAUDRATE");
+        if (baudrate_s == NULL) {
+            serial_baudrate = 115200;
+        } else {
+            serial_baudrate = strtoui(baudrate_s, NULL, 10);
+            if (serial_baudrate == 0 || serial_baudrate > 115200) {
+                serial_baudrate = 115200;
+            }
+        }
+    }
+#endif
+
     char *hash_mismatch_panic_str = config_get_value(NULL, 0, "HASH_MISMATCH_PANIC");
     hash_mismatch_panic = hash_mismatch_panic_str == NULL || strcmp(hash_mismatch_panic_str, "yes") == 0;
 
@@ -828,7 +847,7 @@ noreturn void _menu(bool first_run) {
     if (interface_help_colour_str == NULL) {
         interface_help_colour_str = config_get_value(NULL, 0, "INTERFACE_HELP_COLOR");
     }
-    if (interface_help_colour_str != NULL) {
+    if (interface_help_colour_str != NULL && interface_help_colour_str[0] != '\0') {
         interface_help_colour[3] = interface_help_colour_str[0];
         interface_help_colour_bright[3] = interface_help_colour_str[0];
     }
@@ -954,6 +973,10 @@ noreturn void _menu(bool first_run) {
 
     menu_init_term();
 
+    if (terms[0]->cols < 40 || terms[0]->rows < 16) {
+        panic(false, "Terminal too small (minimum 40x16 required)");
+    }
+
     size_t tree_offset = 0;
 
 refresh:
@@ -998,7 +1021,9 @@ refresh:
                              &selected_menu_entry, &max_tree_len, &max_tree_height);
 
     if (max_entries != 0) {
-        size_t tree_prefix_len = (terms[0]->cols / 2 - DIV_ROUNDUP(max_tree_len, 2)) - 2;
+        size_t half_cols = terms[0]->cols / 2;
+        size_t half_tree = DIV_ROUNDUP(max_tree_len, 2);
+        size_t tree_prefix_len = (half_cols > half_tree + 2) ? (half_cols - half_tree - 2) : 0;
         char *tree_prefix = ext_mem_alloc(tree_prefix_len + 1);
         memset(tree_prefix, ' ', tree_prefix_len);
 
@@ -1011,7 +1036,7 @@ refresh:
         max_entries = print_tree(tree_offset, terms[0]->rows - 8, tree_prefix, 0, 0, selected_entry, menu_tree,
                                  &selected_menu_entry, NULL, NULL);
 
-        pmm_free(tree_prefix, tree_prefix_len);
+        pmm_free(tree_prefix, tree_prefix_len + 1);
     }
 
     {
@@ -1067,8 +1092,8 @@ refresh:
         for (size_t i = timeout; i; i--) {
             set_cursor_pos_helper(0, terms[0]->rows - 1);
             FOR_TERM(TERM->scroll_enabled = false);
-            print("\e[2K%sBooting automatically in %s%u%s, press any key to stop the countdown...\e[0m",
-                  interface_help_colour, interface_help_colour_bright, i, interface_help_colour);
+            print("\e[2K%sBooting automatically in %s%U%s, press any key to stop the countdown...\e[0m",
+                  interface_help_colour, interface_help_colour_bright, (uint64_t)i, interface_help_colour);
             FOR_TERM(TERM->scroll_enabled = true);
             FOR_TERM(TERM->double_buffer_flush(TERM));
             if ((c = pit_sleep_and_quit_on_keypress(1))) {
