@@ -112,6 +112,11 @@ noreturn void chainload(char *config, char *cmdline) {
     }
 
     struct volume *p = volume_get_by_coord(false, drive, part);
+    if (p == NULL && config_get_value(config, 0, "GPT_GUID") == NULL
+                  && config_get_value(config, 0, "GPT_UUID") == NULL
+                  && config_get_value(config, 0, "MBR_ID") == NULL) {
+        panic(true, "bios: Specified drive/partition not found");
+    }
 
     char *gpt_guid_s = config_get_value(config, 0, "GPT_GUID");
     if (gpt_guid_s == NULL) {
@@ -324,7 +329,7 @@ noreturn void chainload(char *config, char *cmdline) {
                             (EFI_DEVICE_PATH *)memdev_path,
                             ptr, image_size, &new_handle);
     if (status) {
-        panic(false, "efi: LoadImage failure (%x)", status);
+        panic(false, "efi: LoadImage failure (%X)", (uint64_t)status);
     }
 
     EFI_GUID loaded_img_prot_guid = EFI_LOADED_IMAGE_PROTOCOL_GUID;
@@ -333,7 +338,7 @@ noreturn void chainload(char *config, char *cmdline) {
     status = gBS->HandleProtocol(new_handle, &loaded_img_prot_guid,
                                  (void **)&new_handle_loaded_image);
     if (status) {
-        panic(false, "efi: HandleProtocol failure (%x)", status);
+        panic(false, "efi: HandleProtocol failure (%X)", (uint64_t)status);
     }
 
     if (efi_part_handle != 0) {
@@ -342,7 +347,7 @@ noreturn void chainload(char *config, char *cmdline) {
 
     new_handle_loaded_image->FilePath = efi_file_path;
 
-    new_handle_loaded_image->LoadOptionsSize = cmdline_len * sizeof(CHAR16);
+    new_handle_loaded_image->LoadOptionsSize = (cmdline_len + 1) * sizeof(CHAR16);
     new_handle_loaded_image->LoadOptions = new_cmdline;
 
     bli_on_boot();
@@ -353,7 +358,7 @@ noreturn void chainload(char *config, char *cmdline) {
 
     status = gBS->Exit(efi_image_handle, exit_status, exit_data_size, exit_data);
     if (status) {
-        panic(false, "efi: Exit failure (%x)", status);
+        panic(false, "efi: Exit failure (%X)", (uint64_t)status);
     }
 
     __builtin_unreachable();
