@@ -385,18 +385,27 @@ noreturn void linux_load(char *config, char *cmdline) {
     setup_header->loadflags &= ~(1 << 5);     // print early messages
 
     // load kernel
-    uintptr_t kernel_load_addr = 0x100000;
+    size_t kernel_data_size = kernel_file->size - real_mode_code_size;
+    size_t kernel_alloc_size = kernel_data_size;
+    if (setup_header->version >= 0x20a && setup_header->init_size > kernel_alloc_size) {
+        kernel_alloc_size = setup_header->init_size;
+    }
+    uintptr_t kernel_align = 0x100000;
+    if (setup_header->version >= 0x205 && setup_header->kernel_alignment > kernel_align) {
+        kernel_align = setup_header->kernel_alignment;
+    }
+    uintptr_t kernel_load_addr = ALIGN_UP(0x100000, kernel_align);
     for (;;) {
         if (memmap_alloc_range(kernel_load_addr,
-                ALIGN_UP(kernel_file->size - real_mode_code_size, 4096),
+                ALIGN_UP(kernel_alloc_size, 4096),
                 MEMMAP_BOOTLOADER_RECLAIMABLE, MEMMAP_USABLE, false, false, false))
             break;
 
-        if (kernel_load_addr == 0xfff00000) {
+        if (kernel_load_addr >= 0xfff00000) {
             panic(true, "linux: Failed to allocate memory for kernel");
         }
 
-        kernel_load_addr += 0x100000;
+        kernel_load_addr += kernel_align;
     }
 
     fread(kernel_file, (void *)kernel_load_addr, real_mode_code_size, kernel_file->size - real_mode_code_size);
@@ -432,7 +441,9 @@ noreturn void linux_load(char *config, char *cmdline) {
         if ((module = uri_open(module_path)) == NULL)
             panic(true, "linux: Failed to open module with path `%s`. Is the path correct?", module_path);
 
-        size_of_all_modules += module->size;
+        if (__builtin_add_overflow(size_of_all_modules, module->size, &size_of_all_modules)) {
+            panic(true, "linux: Total module size overflow");
+        }
 
         modules[i] = module;
     }
@@ -442,9 +453,12 @@ noreturn void linux_load(char *config, char *cmdline) {
     if (setup_header->version <= 0x202 || setup_header->initrd_addr_max == 0) {
         modules_mem_base = 0x38000000;
     } else {
-        modules_mem_base = setup_header->initrd_addr_max + 1;
+        modules_mem_base = (uintptr_t)setup_header->initrd_addr_max + 1;
     }
 
+    if (size_of_all_modules > modules_mem_base) {
+        panic(true, "linux: Total module size exceeds available address space");
+    }
     modules_mem_base -= size_of_all_modules;
     modules_mem_base = ALIGN_DOWN(modules_mem_base, 0x100000);
 
@@ -566,6 +580,7 @@ set_textmode:;
 
 #if defined (BIOS)
         screen_info->orig_video_isVGA = VIDEO_TYPE_VLFB;
+        screen_info->lfb_size = DIV_ROUNDUP(screen_info->lfb_size, 65536);
 #elif defined (UEFI)
         screen_info->orig_video_isVGA = VIDEO_TYPE_EFI;
 #endif
