@@ -134,6 +134,10 @@ noreturn void multiboot1_load(char *config, char *cmdline) {
             bss_size = header.bss_end_addr - bss_addr;
         }
 
+        if (load_src + load_size > kernel_file_size) {
+            panic(true, "multiboot1: load_src + load_size exceeds kernel file size");
+        }
+
         size_t full_size = load_size + bss_size;
 
         void *elsewhere = ext_mem_alloc(full_size);
@@ -152,14 +156,14 @@ noreturn void multiboot1_load(char *config, char *cmdline) {
 
         switch (bits) {
             case 32:
-                if (!elf32_load_elsewhere(kernel, &entry_point, &ranges))
+                if (!elf32_load_elsewhere(kernel, kernel_file_size, &entry_point, &ranges))
                     panic(true, "multiboot1: ELF32 load failure");
 
                 section_hdr_info = elf32_section_hdr_info(kernel);
                 section_hdr_info_valid = true;
                 break;
             case 64: {
-                if (!elf64_load_elsewhere(kernel, &entry_point, &ranges))
+                if (!elf64_load_elsewhere(kernel, kernel_file_size, &entry_point, &ranges))
                     panic(true, "multiboot1: ELF64 load failure");
 
                 section_hdr_info = elf64_section_hdr_info(kernel);
@@ -336,7 +340,7 @@ noreturn void multiboot1_load(char *config, char *cmdline) {
             fclose(f);
 
             if (verbose) {
-                print("multiboot1: Requested module %u:\n", i);
+                print("multiboot1: Requested module %u:\n", (uint32_t)i);
                 print("            Path:   %s\n", module_path);
                 print("            String: \"%s\"\n", module_cmdline ?: "");
                 print("            Begin:  %x\n", m->begin);
