@@ -118,6 +118,15 @@ noreturn void multiboot2_load(char *config, char* cmdline) {
         panic(true, "multiboot2: Header checksum is invalid");
     }
 
+    if (header->architecture != MULTIBOOT_ARCHITECTURE_I386) {
+        panic(true, "multiboot2: Unsupported architecture %u (expected i386)", header->architecture);
+    }
+
+    size_t header_offset_in_file = (uintptr_t)header - (uintptr_t)kernel;
+    if (header->header_length > kernel_file_size - header_offset_in_file) {
+        panic(true, "multiboot2: Header length exceeds kernel file size");
+    }
+
     struct multiboot_header_tag_address *addresstag = NULL;
     struct multiboot_header_tag_framebuffer *fbtag = NULL;
 
@@ -171,9 +180,6 @@ noreturn void multiboot2_load(char *config, char* cmdline) {
                         case MULTIBOOT_TAG_TYPE_EFI32_IH:
                         case MULTIBOOT_TAG_TYPE_EFI64:
                         case MULTIBOOT_TAG_TYPE_EFI64_IH:
-                            if (is_required) {
-                                panic(true, "multiboot2: EFI entry point tag handling is required but Limine does not support it");
-                            }
                             break;
                         #endif
                         case MULTIBOOT_TAG_TYPE_FRAMEBUFFER:
@@ -190,6 +196,9 @@ noreturn void multiboot2_load(char *config, char* cmdline) {
                         case MULTIBOOT_TAG_TYPE_ELF_SECTIONS:
                             is_elf_info_requested = is_required;
                             break;
+                        case MULTIBOOT_TAG_TYPE_LOAD_BASE_ADDR:
+                        case MULTIBOOT_TAG_TYPE_NETWORK:
+                            break;
                         default:
                             if (is_required)
                                 panic(true, "multiboot2: Requested tag `%d` which is not supported", r);
@@ -222,7 +231,12 @@ noreturn void multiboot2_load(char *config, char* cmdline) {
             }
             // We always align the modules ;^)
             case MULTIBOOT_HEADER_TAG_MODULE_ALIGN:
+                break;
             case MULTIBOOT_HEADER_TAG_EFI_BS:
+                print("multiboot2: Warning: EFI_BS tag requested but not supported\n");
+                if (is_required) {
+                    panic(true, "multiboot2: EFI_BS tag is required but not supported");
+                }
                 break;
 
             case MULTIBOOT_HEADER_TAG_RELOCATABLE: {
@@ -303,6 +317,10 @@ noreturn void multiboot2_load(char *config, char* cmdline) {
             bss_size = addresstag->bss_end_addr - bss_addr;
         }
 
+        if (load_src + load_size > kernel_file_size) {
+            panic(true, "multiboot2: load_src + load_size exceeds kernel file size");
+        }
+
         size_t full_size = load_size + bss_size;
 
         void *elsewhere = ext_mem_alloc(full_size);
@@ -324,14 +342,14 @@ noreturn void multiboot2_load(char *config, char* cmdline) {
 
         switch (bits) {
             case 32:
-                if (!elf32_load_elsewhere(kernel, &e, &ranges))
+                if (!elf32_load_elsewhere(kernel, kernel_file_size, &e, &ranges))
                     panic(true, "multiboot2: ELF32 load failure");
 
                 section_hdr_info = elf32_section_hdr_info(kernel);
                 section_hdr_info_valid = true;
                 break;
             case 64: {
-                if (!elf64_load_elsewhere(kernel, &e, &ranges))
+                if (!elf64_load_elsewhere(kernel, kernel_file_size, &e, &ranges))
                     panic(true, "multiboot2: ELF64 load failure");
 
                 section_hdr_info = elf64_section_hdr_info(kernel);
@@ -350,28 +368,32 @@ noreturn void multiboot2_load(char *config, char* cmdline) {
     int64_t reloc_slide = 0;
 
     if (has_reloc_header) {
+        if (reloc_tag.align == 0) {
+            panic(true, "multiboot2: Relocatable tag has align=0");
+        }
+
         bool reloc_ascend;
         uint64_t relocated_base;
 
         switch (reloc_tag.preference) {
             default:
-            case 0: case 1: case 2: // Prefer lowest to highest
+            case 0: case 1: // No preference / prefer lowest
                 reloc_ascend = true;
                 relocated_base = ALIGN_UP(reloc_tag.min_addr, reloc_tag.align);
                 if (relocated_base + ranges->length > reloc_tag.max_addr) {
                     goto reloc_fail;
                 }
                 break;
-/*
-            XXX: We do not support this properly, hence ignore, for now.
-            case 2: // Prefer highest to lowest
+            case 2: // Prefer highest
                 reloc_ascend = false;
+                if (ranges->length > reloc_tag.max_addr) {
+                    goto reloc_fail;
+                }
                 relocated_base = ALIGN_DOWN(reloc_tag.max_addr - ranges->length, reloc_tag.align);
                 if (relocated_base < reloc_tag.min_addr) {
                     goto reloc_fail;
                 }
                 break;
-*/
         }
 
         for (;;) {
@@ -385,10 +407,10 @@ noreturn void multiboot2_load(char *config, char* cmdline) {
                     goto reloc_fail;
                 }
             } else {
-                relocated_base -= reloc_tag.align;
-                if (relocated_base < reloc_tag.min_addr) {
+                if (relocated_base - reloc_tag.min_addr < reloc_tag.align) {
                     goto reloc_fail;
                 }
+                relocated_base -= reloc_tag.align;
             }
         }
 
@@ -416,7 +438,7 @@ reloc_fail:
 
         char *module_cmdline = conf_tuple.value2;
         if (!module_cmdline) module_cmdline = "";
-        modules_size += sizeof(struct multiboot_tag_module) + strlen(module_cmdline) + 1;
+        modules_size += ALIGN_UP(sizeof(struct multiboot_tag_module) + strlen(module_cmdline) + 1, MULTIBOOT_TAG_ALIGN);
     }
 
     struct smbios_entry_point_32* smbios_entry_32 = NULL;
@@ -427,9 +449,9 @@ reloc_fail:
     uint32_t smbios_tag_size = 0;
 
     if (smbios_entry_32 != NULL)
-        smbios_tag_size += sizeof(struct multiboot_tag_smbios) + smbios_entry_32->length;
+        smbios_tag_size += ALIGN_UP(sizeof(struct multiboot_tag_smbios) + smbios_entry_32->length, MULTIBOOT_TAG_ALIGN);
     if (smbios_entry_64 != NULL)
-        smbios_tag_size += sizeof(struct multiboot_tag_smbios) + smbios_entry_64->length;
+        smbios_tag_size += ALIGN_UP(sizeof(struct multiboot_tag_smbios) + smbios_entry_64->length, MULTIBOOT_TAG_ALIGN);
 
     size_t mb2_info_size = get_multiboot2_info_size(
         cmdline,
@@ -499,11 +521,20 @@ reloc_fail:
 
         int bits = elf_bits(kernel);
 
+        if ((bits == 64 && section_hdr_info.section_entry_size < sizeof(struct elf64_shdr)) ||
+            (bits == 32 && section_hdr_info.section_entry_size < sizeof(struct elf32_shdr))) {
+            panic(true, "multiboot2: ELF section entry size too small");
+        }
+
         for (size_t i = 0; i < section_hdr_info.num; i++) {
             if (bits == 64)  {
                 struct elf64_shdr *shdr = (void *)tag->sections + i * section_hdr_info.section_entry_size;
 
-                if (shdr->sh_addr != 0 || shdr->sh_size == 0) {
+                if (shdr->sh_addr != 0) {
+                    shdr->sh_addr += reloc_slide;
+                    continue;
+                }
+                if (shdr->sh_size == 0) {
                     continue;
                 }
 
@@ -524,7 +555,11 @@ reloc_fail:
             } else {
                 struct elf32_shdr *shdr = (void *)tag->sections + i * section_hdr_info.section_entry_size;
 
-                if (shdr->sh_addr != 0 || shdr->sh_size == 0) {
+                if (shdr->sh_addr != 0) {
+                    shdr->sh_addr += (int32_t)reloc_slide;
+                    continue;
+                }
+                if (shdr->sh_size == 0) {
                     continue;
                 }
 
@@ -602,7 +637,7 @@ reloc_fail:
         fclose(f);
 
         if (verbose) {
-            print("multiboot2: Requested module %u:\n", i);
+            print("multiboot2: Requested module %u:\n", (uint32_t)i);
             print("            Path:   %s\n", module_path);
             print("            String: \"%s\"\n", module_cmdline ?: "");
             print("            Begin:  %x\n", module_tag->mod_start);
@@ -669,7 +704,6 @@ reloc_fail:
         struct multiboot_tag_framebuffer *tag = (struct multiboot_tag_framebuffer *)(mb2_info + info_idx);
 
         tag->common.type = MULTIBOOT_TAG_TYPE_FRAMEBUFFER;
-        tag->common.size = sizeof(struct multiboot_tag_framebuffer);
 
         term_notready();
 
@@ -712,6 +746,7 @@ textmode:
                 tag->common.framebuffer_height = 25;
                 tag->common.framebuffer_bpp = 16;
                 tag->common.framebuffer_type = MULTIBOOT_FRAMEBUFFER_TYPE_EGA_TEXT;
+                tag->common.size = sizeof(struct multiboot_tag_framebuffer_common);
 #elif defined (UEFI)
                 if (is_framebuffer_required) {
                     panic(true, "multiboot2: Failed to set video mode");
@@ -726,6 +761,7 @@ textmode:
                 tag->common.framebuffer_height = fbs[0].framebuffer_height;
                 tag->common.framebuffer_bpp = fbs[0].framebuffer_bpp;
                 tag->common.framebuffer_type = MULTIBOOT_FRAMEBUFFER_TYPE_RGB; // We only support RGB for VBE
+                tag->common.size = sizeof(struct multiboot_tag_framebuffer);
 
                 tag->framebuffer_red_field_position = fbs[0].red_mask_shift;
                 tag->framebuffer_red_mask_size = fbs[0].red_mask_size;
