@@ -66,14 +66,14 @@ void *memcpy(void *, const void *, size_t);
 
 static uint8_t bump_alloc_pool[FLANTERM_FB_BUMP_ALLOC_POOL_SIZE];
 static size_t bump_alloc_ptr = 0;
+static bool bump_alloc_base_offset_added = false;
 
 static void *bump_alloc(size_t s) {
-    static bool base_offset_added = false;
-    if (!base_offset_added) {
+    if (!bump_alloc_base_offset_added) {
         if ((uintptr_t)bump_alloc_pool & 0xf) {
             bump_alloc_ptr += 0x10 - ((uintptr_t)bump_alloc_pool & 0xf);
         }
-        base_offset_added = true;
+        bump_alloc_base_offset_added = true;
     }
 
     if ((s & 0xf) != 0) {
@@ -446,7 +446,17 @@ static ALWAYS_INLINE uint32_t convert_colour(struct flanterm_context *_ctx, uint
     uint32_t r = (colour >> 16) & 0xff;
     uint32_t g = (colour >> 8) & 0xff;
     uint32_t b =  colour & 0xff;
-    return (r << ctx->red_mask_shift) | (g << ctx->green_mask_shift) | (b << ctx->blue_mask_shift);
+    uint32_t ret = (r << ctx->red_mask_shift) | (g << ctx->green_mask_shift) | (b << ctx->blue_mask_shift);
+    if (ctx->red_mask_size > 8) {
+        ret |= (r >> (16 - ctx->red_mask_size)) << (ctx->red_mask_shift + 8);
+    }
+    if (ctx->green_mask_size > 8) {
+        ret |= (g >> (16 - ctx->green_mask_size)) << (ctx->green_mask_shift + 8);
+    }
+    if (ctx->blue_mask_size > 8) {
+        ret |= (b >> (16 - ctx->blue_mask_size)) << (ctx->blue_mask_shift + 8);
+    }
+    return ret;
 }
 
 static void flanterm_fb_save_state(struct flanterm_context *_ctx) {
@@ -470,6 +480,12 @@ static void flanterm_fb_swap_palette(struct flanterm_context *_ctx) {
     uint32_t tmp = ctx->text_bg;
     ctx->text_bg = ctx->text_fg;
     ctx->text_fg = tmp;
+    if (ctx->text_fg == 0xffffffff) {
+        ctx->text_fg = ctx->default_bg;
+    }
+    if (ctx->text_bg == ctx->default_bg) {
+        ctx->text_bg = 0xffffffff;
+    }
 }
 
 static void plot_char_scaled_canvas(struct flanterm_context *_ctx, struct flanterm_fb_char *c, size_t x, size_t y) {
@@ -541,7 +557,7 @@ static void plot_char_scaled_uncanvas(struct flanterm_context *_ctx, struct flan
     uint32_t default_bg = ctx->default_bg;
 
     uint32_t bg = c->bg == 0xffffffff ? default_bg : c->bg;
-    uint32_t fg = c->fg == 0xffffffff ? default_bg : c->fg;
+    uint32_t fg = c->fg == 0xffffffff ? ctx->default_fg : c->fg;
 
     x = ctx->offset_x + x * ctx->glyph_width;
     y = ctx->offset_y + y * ctx->glyph_height;
@@ -655,7 +671,7 @@ static void plot_char_unscaled_uncanvas(struct flanterm_context *_ctx, struct fl
     uint32_t default_bg = ctx->default_bg;
 
     uint32_t bg = c->bg == 0xffffffff ? default_bg : c->bg;
-    uint32_t fg = c->fg == 0xffffffff ? default_bg : c->fg;
+    uint32_t fg = c->fg == 0xffffffff ? ctx->default_fg : c->fg;
 
     x = ctx->offset_x + x * ctx->glyph_width;
     y = ctx->offset_y + y * ctx->glyph_height;
@@ -720,6 +736,9 @@ static void push_to_queue(struct flanterm_context *_ctx, struct flanterm_fb_char
         if (compare_char(&ctx->grid[i], c)) {
             return;
         }
+        if (ctx->queue_i == _ctx->rows * _ctx->cols) {
+            return;
+        }
         q = &ctx->queue[ctx->queue_i++];
         q->x = x;
         q->y = y;
@@ -732,11 +751,10 @@ static void push_to_queue(struct flanterm_context *_ctx, struct flanterm_fb_char
 static void flanterm_fb_revscroll(struct flanterm_context *_ctx) {
     struct flanterm_fb_context *ctx = (void *)_ctx;
 
-    for (size_t i = (_ctx->scroll_bottom_margin - 1) * _ctx->cols - 1;
-         i >= _ctx->scroll_top_margin * _ctx->cols; i--) {
-        if (i == (size_t)-1) {
-            break;
-        }
+    size_t start = _ctx->scroll_top_margin * _ctx->cols;
+    size_t end = (_ctx->scroll_bottom_margin - 1) * _ctx->cols;
+    for (size_t i = end; i > start; ) {
+        i--;
         struct flanterm_fb_char *c;
         struct flanterm_fb_queue_item *q = ctx->map[i];
         if (q != NULL) {
@@ -803,14 +821,14 @@ static void flanterm_fb_set_cursor_pos(struct flanterm_context *_ctx, size_t x,
     struct flanterm_fb_context *ctx = (void *)_ctx;
 
     if (x >= _ctx->cols) {
-        if ((int)x < 0) {
+        if (x > SIZE_MAX / 2) {
             x = 0;
         } else {
             x = _ctx->cols - 1;
         }
     }
     if (y >= _ctx->rows) {
-        if ((int)y < 0) {
+        if (y > SIZE_MAX / 2) {
             y = 0;
         } else {
             y = _ctx->rows - 1;
@@ -964,22 +982,26 @@ static void flanterm_fb_double_buffer_flush(struct flanterm_context *_ctx) {
     ctx->queue_i = 0;
 
     if (ctx->flush_callback) {
-        ctx->flush_callback(ctx->framebuffer, ctx->pitch * ctx->height);
+        ctx->flush_callback(ctx->framebuffer, ctx->pitch * ctx->phys_height);
     }
 }
 
 static void flanterm_fb_raw_putchar(struct flanterm_context *_ctx, uint8_t c) {
     struct flanterm_fb_context *ctx = (void *)_ctx;
 
-    if (ctx->cursor_x >= _ctx->cols && (ctx->cursor_y < _ctx->scroll_bottom_margin - 1 || _ctx->scroll_enabled)) {
-        ctx->cursor_x = 0;
-        ctx->cursor_y++;
-        if (ctx->cursor_y == _ctx->scroll_bottom_margin) {
-            ctx->cursor_y--;
-            flanterm_fb_scroll(_ctx);
-        }
-        if (ctx->cursor_y >= _ctx->cols) {
-            ctx->cursor_y = _ctx->cols - 1;
+    if (ctx->cursor_x >= _ctx->cols) {
+        if (_ctx->wrap_enabled && (ctx->cursor_y < _ctx->scroll_bottom_margin - 1 || _ctx->scroll_enabled)) {
+            ctx->cursor_x = 0;
+            ctx->cursor_y++;
+            if (ctx->cursor_y == _ctx->scroll_bottom_margin) {
+                ctx->cursor_y--;
+                flanterm_fb_scroll(_ctx);
+            }
+            if (ctx->cursor_y >= _ctx->rows) {
+                ctx->cursor_y = _ctx->rows - 1;
+            }
+        } else {
+            ctx->cursor_x = _ctx->cols - 1;
         }
     }
 
@@ -1034,7 +1056,7 @@ static void flanterm_fb_full_refresh(struct flanterm_context *_ctx) {
     }
 
     if (ctx->flush_callback) {
-        ctx->flush_callback(ctx->framebuffer, ctx->pitch * ctx->height);
+        ctx->flush_callback(ctx->framebuffer, ctx->pitch * ctx->phys_height);
     }
 }
 
@@ -1045,6 +1067,7 @@ static void flanterm_fb_deinit(struct flanterm_context *_ctx, void (*_free)(void
 #ifndef FLANTERM_FB_DISABLE_BUMP_ALLOC
         if (bump_allocated_instance == true) {
             bump_alloc_ptr = 0;
+            bump_alloc_base_offset_added = false;
             bump_allocated_instance = false;
         }
 #endif
@@ -1080,6 +1103,8 @@ struct flanterm_context *flanterm_fb_init(
     size_t margin,
     int rotation
 ) {
+    size_t phys_height = height;
+
     if (rotation == FLANTERM_FB_ROTATE_90 || rotation == FLANTERM_FB_ROTATE_270) {
         size_t tmp = width;
         width = height;
@@ -1205,9 +1230,11 @@ struct flanterm_context *flanterm_fb_init(
     ctx->framebuffer = (void *)framebuffer;
     ctx->width = width;
     ctx->height = height;
+    ctx->phys_height = phys_height;
     ctx->pitch = pitch;
 
-#define FONT_BYTES ((font_width * font_height * FLANTERM_FB_FONT_GLYPHS) / 8)
+    // VGA fonts are always one byte per scanline regardless of font_width
+#define FONT_BYTES (font_height * FLANTERM_FB_FONT_GLYPHS)
 
     if (font != NULL) {
         ctx->font_width = font_width;
