@@ -49,10 +49,14 @@ struct flanterm_context {
     bool autoflush;
     bool cursor_enabled;
     bool scroll_enabled;
+    bool wrap_enabled;
+    bool origin_mode;
     bool control_sequence;
     bool escape;
     bool osc;
     bool osc_escape;
+    size_t osc_buf_i;
+    uint8_t osc_buf[256];
     bool rrr;
     bool discard_next;
     bool bold;
@@ -75,11 +79,15 @@ struct flanterm_context {
     size_t scroll_top_margin;
     size_t scroll_bottom_margin;
     uint32_t esc_values[FLANTERM_MAX_ESC_VALUES];
-    uint64_t oob_output;
+    uint8_t last_printed_char;
+    bool last_was_graphic;
     bool saved_state_bold;
     bool saved_state_bg_bold;
     bool saved_state_reverse_video;
+    bool saved_state_origin_mode;
+    bool saved_state_wrap_enabled;
     size_t saved_state_current_charset;
+    uint8_t saved_state_charsets[2];
     size_t saved_state_current_primary;
     size_t saved_state_current_bg;
 
